@@ -12,25 +12,52 @@ use File::Compare ();
 use Data::Dumper ();
 use IO::File ();
 
+#################### Constructors ###########################
 sub new {
-  my $package = shift;
-  my %input = @_;
+  my $self = shift()->_construct(@_);
+
+  $self->cull_args(@ARGV);
+  
+  die "Too early to specify a build action '$self->{action}'.  Do 'Build $self->{action}' instead.\n"
+    if $self->{action};
+
+  $self->_set_install_paths;
+  $self->dist_name;
+  $self->check_manifest;
+  $self->check_prereq;
+  $self->dist_version;
+
+  return $self;
+}
+
+sub resume {
+  my $self = shift()->_construct(@_);
+  
+  $self->read_config;
+  
+  my $perl = $self->find_perl_interpreter;
+  warn(" * WARNING: Configuration was initially created with '$self->{properties}{perl}',\n".
+       "   but we are now using '$perl'.\n")
+    unless $perl eq $self->{properties}{perl};
+  
+  $self->cull_args(@ARGV);
+  $self->{action} ||= 'build';
+  
+  return $self;
+}
+
+sub current {
+  # hmm, wonder what the right thing to do here is
+  local @ARGV;
+  return shift()->resume;
+}
+
+sub _construct {
+  my ($package, %input) = @_;
 
   my $args   = delete $input{args}   || {};
   my $config = delete $input{config} || {};
 
-  my ($action, $cmd_args) = __PACKAGE__->cull_args(@ARGV);
-  die "Too early to specify a build action '$action'.  Do 'Build $action' instead.\n"
-    if $action;
-
-  # Extract our 'properties' from $cmd_args, the rest are put in 'args'
-  my $cmd_properties = {};
-  foreach my $key (keys %$cmd_args) {
-    $cmd_properties->{$key} = delete $cmd_args->{$key} if __PACKAGE__->valid_property($key);
-  }
-
-  my $cmd_config = delete $cmd_args->{config};
-
   # The following warning could be unnecessary if the user is running
   # an embedded perl, but there aren't too many of those around, and
   # embedded perls aren't usually used to install modules, and the
@@ -39,31 +66,29 @@ sub new {
   my $perl = $package->find_perl_interpreter
     or warn "Warning: Can't locate your perl binary";
 
-  # 'args' are arbitrary user args.
-  # 'config' is Config.pm and its overridden values.
-  # 'properties' is stuff Module::Build needs in order to work.  They get saved in _build/.
-  # Anything else in $self doesn't get saved.
-
   my $self = bless {
-		    args => {%$args, %$cmd_args},
-		    config => {%Config, %$config, %$cmd_config},
+		    args => {%$args},
+		    config => {%Config, %$config},
 		    properties => {
-				   build_script => 'Build',
-				   base_dir => $package->cwd,
-				   config_dir => '_build',
-				   requires => {},
-				   recommends => {},
-				   build_requires => {},
-				   conflicts => {},
-				   perl => $perl,
-				   install_types => [qw(lib arch script)],
-				   installdirs => 'site',
-				   include_dirs => [],
+				   build_script    => 'Build',
+				   base_dir        => $package->cwd,
+				   config_dir      => '_build',
+				   blib            => 'blib',
+				   requires        => {},
+				   recommends      => {},
+				   build_requires  => {},
+				   conflicts       => {},
+				   perl            => $perl,
+				   install_types   => [qw( lib arch script bindoc libdoc )],
+				   installdirs     => 'site',
+				   include_dirs    => [],
 				   %input,
-				   %$cmd_properties,
 				  },
 		   }, $package;
+
   my $p = $self->{properties};
+  $p->{bindoc_dirs} ||= [ "$p->{blib}/script" ];
+  $p->{libdoc_dirs} ||= [ "$p->{blib}/lib", "$p->{blib}/arch" ];
 
   # Synonyms
   $p->{requires} = delete $p->{prereq} if exists $p->{prereq};
@@ -72,15 +97,11 @@ sub new {
   $self->add_to_cleanup( @{delete $p->{add_to_cleanup}} )
     if $p->{add_to_cleanup};
   
-  $self->_set_install_paths;
-  $self->dist_name;
-  $self->check_manifest;
-  $self->check_prereq;
-  $self->dist_version;
-
   return $self;
 }
 
+################## End constructors #########################
+
 sub _set_install_paths {
   my $self = shift;
   my $c = $self->{config};
@@ -100,16 +121,16 @@ sub _set_install_paths {
 		arch    => $c->{installsitearch},
 		bin     => $c->{installsitebin} || $c->{installbin},
 		script  => $c->{installsitescript} || $c->{installsitebin} || $c->{installscript},
-		bindoc  => $c->{installsiteman1dir},
-		libdoc  => $c->{installsiteman3dir},
+		bindoc  => $c->{installsiteman1dir} || $c->{installman1dir},
+		libdoc  => $c->{installsiteman3dir} || $c->{installman3dir},
 	       },
      vendor => {
 		lib     => $c->{installvendorlib},
 		arch    => $c->{installvendorarch},
 		bin     => $c->{installvendorbin} || $c->{installbin},
 		script  => $c->{installvendorscript} || $c->{installvendorbin} || $c->{installscript},
-		bindoc  => $c->{installvendorman1dir},
-		libdoc  => $c->{installvendorman3dir},
+		bindoc  => $c->{installvendorman1dir} || $c->{installman1dir},
+		libdoc  => $c->{installvendorman3dir} || $c->{installman3dir},
 	       },
     };
 }
@@ -128,6 +149,7 @@ sub find_perl_interpreter {
 }
 
 sub base_dir { shift()->{properties}{base_dir} }
+sub installdirs { shift()->{properties}{installdirs} }
 
 sub prompt {
   my $self = shift;
@@ -173,30 +195,15 @@ sub y_n {
   }
 }
 
-sub resume {
-  my $package = shift;
-  my $self = bless {@_}, $package;
+sub notes {
+  my $self = shift;
+  return $self->_persistent_hash_read('notes') unless @_;
   
-  $self->read_config;
+  my $key = shift;
+  return $self->_persistent_hash_read('notes', $key) unless @_;
   
-  my $perl = $self->find_perl_interpreter;
-  warn(" * WARNING: Configuration was initially created with '$self->{properties}{perl}',\n".
-       "   but we are now using '$perl'.\n")
-    unless $perl eq $self->{properties}{perl};
-  
-
-  ($self->{action}, my $args) = $self->cull_args(@ARGV);
-  $self->{action} ||= 'build';
-  
-  # Extract our 'properties' from $args
-  my %p;
-  foreach my $key (keys %$args) {
-    $p{$key} = delete $args->{$key} if __PACKAGE__->valid_property($key);
-  }
-  $self->{args} = {%{$self->{args}}, %$args};
-  $self->{properties} = {%{$self->{properties}}, %p};
-
-  return $self;
+  my $value = shift;
+  return $self->_persistent_hash_write('notes', { $key => $value });
 }
 
 {
@@ -217,12 +224,15 @@ sub resume {
        PL_files
        scripts
        script_files
+       test_files
        perl
        config_dir
+       blib
        build_script
        install_types
        install_sets
        install_path
+       install_base
        installdirs
        destdir
        debugger
@@ -232,9 +242,22 @@ sub resume {
        create_makefile_pl
        pollute
        include_dirs
+       bindoc_dirs
+       libdoc_dirs
       );
 
   sub valid_property { exists $valid_properties{$_[1]} }
+
+  # Create an accessor for each property that doesn't already have one
+  foreach my $property (keys %valid_properties) {
+      next if __PACKAGE__->can($property);
+      no strict 'refs';
+      *{$property} = sub {
+          my $self = shift;
+          $self->{properties}{$property} = shift if @_;
+          return $self->{properties}{$property};
+      };
+  }
 }
 
 # XXX Problem - if Module::Build is loaded from a different directory,
@@ -362,74 +385,118 @@ sub find_module_by_name {
   return;
 }
 
+sub _next_code_line {
+  my ($self, $fh, $pat) = @_;
+  my $inpod = 0;
+  
+  local $_;
+  while (<$fh>) {
+    $inpod = /^=(?!cut)/ ? 1 : /^=cut/ ? 0 : $inpod;
+    next if $inpod || /^\s*#/;
+    return wantarray ? ($_, /$pat/) : $_
+      if $_ =~ $pat;
+  }
+  return;
+}
+
 sub version_from_file {
   my ($self, $file) = @_;
 
   # Some of this code came from the ExtUtils:: hierarchy.
   my $fh = IO::File->new($file) or die "Can't open '$file' for version: $!";
-  local $_;
-  while (<$fh>) {
-    if ( my ($sigil, $var) = /([\$*])(([\w\:\']*)\bVERSION)\b.*\=/ ) {
-      my $eval = qq{
-		    package Module::Build::Base::_version;
-		    no strict;
+
+  my $match = qr/([\$*])(([\w\:\']*)\bVERSION)\b.*\=/;
+  my ($v_line, $sigil, $var) = $self->_next_code_line($fh, $match) or return undef;
+
+  my $eval = qq{q#  Hide from _packages_inside()
+		 #; package Module::Build::Base::_version;
+		 no strict;
 		    
-		    local $sigil$var;
-		    \$$var=undef; do {
-		      $_
-		    }; \$$var
-		   };
-      local $^W;
-      return scalar eval $eval;
-    }
+		 local $sigil$var;
+		 \$$var=undef; do {
+		   $v_line
+		 }; \$$var
+		};
+  local $^W;
+  my $result = eval $eval;
+  warn "Error evaling version line '$eval' in $file: $@\n" if $@;
+  return $result;
+}
+
+sub _persistent_hash_write {
+  my ($self, $name, $href) = @_;
+  $href ||= {};
+  my $ph = $self->{phash}{$name} ||= {disk => {}, new => {}};
+  
+  @{$ph->{new}}{ keys %$href } = values %$href;  # Merge
+
+  # Do some optimization to avoid unnecessary writes
+  foreach my $key (keys %{ $ph->{new} }) {
+    next if ref $ph->{new}{$key};
+    next if ref $ph->{disk}{$key} or !exists $ph->{disk}{$key};
+    delete $ph->{new}{$key} if $ph->{new}{$key} eq $ph->{disk}{$key};
   }
-  return undef;
+  
+  if (my $file = $self->config_file($name)) {
+    return if -e $file and !keys %{ $ph->{new} };  # Nothing to do
+    
+    local $Data::Dumper::Terse = 1;
+    my $fh = IO::File->new("> $file") or die "Can't write to $file: $!";
+    @{$ph->{disk}}{ keys %{$ph->{new}} } = values %{$ph->{new}};  # Merge
+    print $fh Data::Dumper::Dumper($ph->{disk});
+    close $fh;
+    
+    $ph->{new} = {};
+  }
+  return $self->_persistent_hash_read($name);
 }
 
-sub _write_cleanup {
+sub _persistent_hash_read {
   my $self = shift;
-  my @to_write = keys %{ $self->{add_to_cleanup} }
-    or return;
-  
-  my $file = $self->config_file('cleanup');
-  my $fh = IO::File->new(">> $file") or die "Can't write to $file: $!";
-  print $fh "$_\n" foreach @to_write;
-  close $fh;
-  
-  @{ $self->{cleanup} }{ @to_write } = ();
-  $self->{add_to_cleanup} = {};
+  my $name = shift;
+  my $ph = $self->{phash}{$name} ||= {disk => {}, new => {}};
+
+  if (@_) {
+    # Return 1 key as a scalar
+    my $key = shift;
+    return $ph->{new}{$key} if exists $ph->{new}{$key};
+    return $ph->{disk}{$key};
+  } else {
+    # Return all data
+    my $out = (keys %{$ph->{new}}
+	       ? {%{$ph->{disk}}, %{$ph->{new}}}
+	       : $ph->{disk});
+    return wantarray ? %$out : $out;
+  }
 }
 
-sub cleanup_is_flushed {
-  my $self = shift;
-  return ! keys %{ $self->{add_to_cleanup} };
+sub _persistent_hash_restore {
+  my ($self, $name) = @_;
+  my $ph = $self->{phash}{$name} ||= {disk => {}, new => {}};
+  
+  my $file = $self->config_file($name) or die "No config file '$name'";
+  my $fh = IO::File->new("< $file") or die "Can't read $file: $!";
+  
+  $ph->{disk} = eval do {local $/; <$fh>};
+  die $@ if $@;
 }
 
 sub add_to_cleanup {
   my $self = shift;
-
-  # $self->{cleanup} contains files that are already written in the
-  # 'cleanup' file.  $self->{add_to_cleanup} is a buffer that we
-  # haven't written yet (and may never write if we don't ever create
-  # the cleanup file).
-  
-  my @new_files = grep {!exists $self->{cleanup}{$_}} @_
-    or return;
-  
-  @{$self->{add_to_cleanup}}{ @new_files } = ();
-  
-  $self->_write_cleanup if $self->config_file('cleanup');
+  my %files = map {$_, 1} @_;
+  $self->_persistent_hash_write('cleanup', \%files);
 }
 
 sub cleanup {
   my $self = shift;
-  return (keys %{$self->{cleanup}}, keys %{$self->{add_to_cleanup}});
+  my $all = $self->_persistent_hash_read('cleanup');
+  return keys %$all;
 }
 
 sub config_file {
   my $self = shift;
-  return unless -d $self->{properties}{config_dir};
-  return File::Spec->catfile($self->{properties}{config_dir}, @_);
+  return unless -d $self->config_dir;
+  return File::Spec->catfile($self->config_dir, @_);
 }
 
 sub read_config {
@@ -441,14 +508,10 @@ sub read_config {
   die if $@;
   ($self->{args}, $self->{config}, $self->{properties}) = @$ref;
   close $fh;
-  
-  my $cleanup_file = $self->config_file('cleanup');
-  $self->{cleanup} = {};
-  if (-e $cleanup_file) {
-    my $fh = IO::File->new($cleanup_file) or die "Can't read '$file': $!";
-    my @files = <$fh>;
-    chomp @files;
-    @{$self->{cleanup}}{@files} = ();
+
+  for ('cleanup', 'notes') {
+    next unless -e $self->config_file($_);
+    $self->_persistent_hash_restore($_);
   }
 }
 
@@ -470,8 +533,8 @@ sub write_config {
   my @items = qw(requires build_requires conflicts recommends);
   print $fh Data::Dumper::Dumper( { map { $_, $self->$_() } @items } );
   close $fh;
-  
-  $self->_write_cleanup;
+
+  $self->_persistent_hash_write('cleanup');
 }
 
 sub requires       { shift()->{properties}{requires} }
@@ -553,9 +616,6 @@ sub check_installed_status {
   my %status = (need => $spec);
   
   if ($modname eq 'perl') {
-    # Check the current perl interpreter
-    # It's much more convenient to use $] here than $^V, but 'man
-    # perlvar' says I'm not supposed to.  Bloody tyrant.
     $status{have} = $self->perl_version;
   
   } elsif (eval { $status{have} = $modname->VERSION }) {
@@ -586,8 +646,7 @@ sub check_installed_status {
     
     next if $op eq '>=' and !$version;  # Module doesn't have to actually define a $VERSION
     
-    unless (eval "\$status{have} $op \$version") {
-      warn $@ if $@;
+    unless ($self->compare_versions( $status{have}, $op, $version )) {
       $status{message} = "Version $status{have} is installed, but we need version $op $version";
       return \%status;
     }
@@ -597,6 +656,21 @@ sub check_installed_status {
   return \%status;
 }
 
+sub compare_versions {
+  my $self = shift;
+  my ($v1, $op, $v2) = @_;
+
+  # for alpha versions - this doesn't cover all cases, but should work for most:
+  $v1 =~ s/_(\d+)\z/$1/;
+  $v2 =~ s/_(\d+)\z/$1/;
+
+  my $eval_str = "\$v1 $op \$v2";
+  my $result   = eval $eval_str;
+  warn "error comparing versions: '$eval_str' $@" if $@;
+
+  return $result;
+}
+
 # I wish I could set $! to a string, but I can't, so I use $@
 sub check_installed_version {
   my ($self, $modname, $spec) = @_;
@@ -628,12 +702,11 @@ sub print_build_script {
   my ($self, $fh) = @_;
   
   my $build_package = ref($self);
-
-  my ($config_dir, $build_script, $base_dir) = 
-    ($self->{properties}{config_dir}, $self->{properties}{build_script}, $self->base_dir);
+  
+  my %q = map {$_, $self->$_()} qw(config_dir base_dir);
 
   my @myINC = @INC;
-  for ($config_dir, $build_script, $base_dir, @myINC) {
+  for (@myINC, values %q) {
     $_ = File::Spec->rel2abs($_);
     s/([\\\'])/\\$1/g;
   }
@@ -645,7 +718,7 @@ $self->{config}{startperl}
 
 BEGIN {
   \$^W = 1;  # Use warnings
-  my \$start_dir = '$base_dir';
+  my \$start_dir = '$q{base_dir}';
   chdir(\$start_dir) or die "Cannot chdir to \$start_dir: \$!";
   \@INC = 
     (
@@ -658,8 +731,7 @@ use $build_package;
 # This should have just enough arguments to be able to bootstrap the rest.
 my \$build = resume $build_package (
   properties => {
-    config_dir => '$config_dir',
-    build_script => '$build_script',
+    config_dir => '$q{config_dir}',
   },
 );
 
@@ -708,14 +780,15 @@ sub check_manifest {
 
 sub dispatch {
   my $self = shift;
+  local $self->{_completed_actions} = {};
 
   if (@_) {
-    (local $self->{action}, my %p) = @_;
+    my ($action, %p) = @_;
     my $args = $p{args} ? delete($p{args}) : {};
     
     local $self->{args} = {%{$self->{args}}, %$args};
     local $self->{properties} = {%{$self->{properties}}, %p};
-    return $self->_call_action($self->{action});
+    return $self->_call_action($action);
   }
 
   die "No build action specified" unless $self->{action};
@@ -724,8 +797,10 @@ sub dispatch {
 
 sub _call_action {
   my ($self, $action) = @_;
-  my $method = "ACTION_$self->{action}";
-  die "No action '$self->{action}' defined" unless $self->can($method);
+  return if $self->{_completed_actions}{$action}++;
+  local $self->{action} = $action;
+  my $method = "ACTION_$action";
+  die "No action '$action' defined" unless $self->can($method);
   return $self->$method();
 }
 
@@ -757,8 +832,12 @@ sub cull_args {
   }
   $args{ARGV} = \@argv;
 
+  # 'config' and 'install_path' are additive by hash key
+  my %additive = (config => 1,
+		  install_path => 1);
+
   # Hashify these parameters
-  for ('config', 'install_path') {
+  for (keys %additive) {
     my %hash;
     $args{$_} ||= [];
     $args{$_} = [ $args{$_} ] unless ref $args{$_};
@@ -770,7 +849,20 @@ sub cull_args {
     $args{$_} = \%hash;
   }
 
-  return ($action, \%args);
+  # Now merge data into $self.
+  $self->{action} = $action if defined $action;
+
+  # Extract our 'properties' from $cmd_args, the rest are put in 'args'.
+  foreach my $key (keys %args) {
+    my $add_to = $self->valid_property($key) ? $self->{properties} : $self->{args};
+
+    if ($additive{$key}) {
+      $add_to->{$key}{$_} = $args{$key}{$_} foreach keys %{$args{$key}};
+    } else {
+      $add_to->{$key} = $args{$key};
+    }
+  }
+
 }
 
 sub super_classes {
@@ -795,11 +887,62 @@ sub known_actions {
     }
   }
 
-  return sort keys %actions;
+  return wantarray ? sort keys %actions : \%actions;
+}
+
+sub _get_action_docs {
+  my ($self, $action, $actions) = @_;
+  $actions ||= $self->known_actions;
+  $@ = '';
+  ($@ = "No known action '$action'\n"), return
+    unless $actions->{$action};
+  
+  my ($files_found, @docs) = (0);
+  foreach my $class ($self->super_classes) {
+    (my $file = $class) =~ s{::}{/}g;
+    $file = $INC{$file . '.pm'} or next;
+    my $fh = IO::File->new("< $file") or next;
+    $files_found++;
+    
+    # Code below modified from /usr/bin/perldoc
+    
+    # Skip to ACTIONS section
+    local $_;
+    while (<$fh>) {
+      last if /^=head1 ACTIONS\s/;
+    }
+    
+    # Look for our action
+    my ($found, $inlist) = (0, 0);
+    while (<$fh>) {
+      if (/^=item\s+\Q$action\E\b/o)  {
+	$found = 1;
+      } elsif (/^=item/) {
+	last if $found > 1 and not $inlist;
+      }
+      next unless $found;
+      push @docs, $_;
+      ++$inlist if /^=over/;
+      --$inlist if /^=back/;
+      ++$found  if /^\w/; # Found descriptive text
+    }
+  }
+  ($@ = "Sorry, couldn't find any documentation to search.\n"), return
+    unless $files_found;
+  ($@ = "Couldn't find any docs for action '$action'.\n"), return
+    unless @docs;
+  
+  return @docs;
 }
 
 sub ACTION_help {
   my ($self) = @_;
+  my $actions = $self->known_actions;
+  
+  if (@{$self->{args}{ARGV}}) {
+    print $self->_get_action_docs($self->{args}{ARGV}[0], $actions), $@;
+    return;
+  }
 
   print <<EOF;
 
@@ -809,15 +952,16 @@ sub ACTION_help {
  Actions defined:
 EOF
 
-  my @actions = $self->known_actions;
   # Flow down columns, not across rows
+  my @actions = sort keys %$actions;
   @actions = map $actions[($_ + ($_ % 2) * @actions) / 2],  0..$#actions;
   
   while (my ($one, $two) = splice @actions, 0, 2) {
     printf("  %-12s                   %-12s\n", $one, $two||'');
   }
-
-  print "\nSee `perldoc Module::Build` for details of the individual actions.\n";
+  
+  print "\nRun `Build help <action>` for details on an individual action.\n";
+  print "See `perldoc Module::Build` for complete documentation.\n";
 }
 
 sub ACTION_test {
@@ -825,7 +969,7 @@ sub ACTION_test {
   my $p = $self->{properties};
   require Test::Harness;
   
-  $self->depends_on('build');
+  $self->depends_on('code');
   
   # Do everything in our power to work with all versions of Test::Harness
   local ($Test::Harness::switches,
@@ -838,15 +982,15 @@ sub ACTION_test {
          $ENV{HARNESS_VERBOSE}) = ($p->{verbose} || 0) x 4;
 
   # Make sure we test the module in blib/
-  local @INC = (File::Spec->catdir($p->{base_dir}, 'blib', 'lib'),
-		File::Spec->catdir($p->{base_dir}, 'blib', 'arch'),
+  local @INC = (File::Spec->catdir($p->{base_dir}, $self->blib, 'lib'),
+		File::Spec->catdir($p->{base_dir}, $self->blib, 'arch'),
 		@INC);
   
   my $tests = $self->test_files;
 
   if (@$tests) {
     # Work around a Test::Harness bug that loses the particular perl we're running under
-    local $^X = $self->{config}{perlpath} unless $Test::Harness::VERSION gt '2.01';
+    local $^X = $p->{perl} unless $Test::Harness::VERSION gt '2.01';
     Test::Harness::runtests(@$tests);
   } else {
     print("No tests defined.\n");
@@ -861,16 +1005,28 @@ sub ACTION_test {
 
 sub test_files {
   my $self = shift;
-  
+  my $p = $self->{properties};
+  if (@_) {
+    $p->{test_files} = (@_ == 1 ? shift : [@_]);
+  }
+
   my @tests;
-  if ($self->{args}{test_files}) {  # XXX use 'properties'
-    @tests = $self->split_like_shell($self->{args}{test_files});
+  if ($p->{test_files}) {
+    @tests = (map { -d $_ ? $self->expand_test_dir($_) : $_ }
+	      map glob,
+	      $self->split_like_shell($p->{test_files}));
   } else {
     # Find all possible tests in t/ or test.pl
     push @tests, 'test.pl'                          if -e 'test.pl';
-    push @tests, @{$self->rscan_dir('t', qr{\.t$})} if -e 't' and -d _;
+    push @tests, $self->expand_test_dir('t')        if -e 't' and -d _;
   }
-  return [sort @tests];
+  return \@tests;
+}
+
+sub expand_test_dir {
+  my ($self, $dir) = @_;
+  return @{$self->rscan_dir($dir, qr{\.t$})} if $self->{properties}{recursive_test_files};
+  return sort glob File::Spec->catfile($dir, "*.t");
 }
 
 sub ACTION_testdb {
@@ -879,12 +1035,12 @@ sub ACTION_testdb {
   $self->depends_on('test');
 }
 
-sub ACTION_build {
+sub ACTION_code {
   my ($self) = @_;
   
   # All installable stuff gets created in blib/ .
   # Create blib/arch to keep blib.pm happy
-  my $blib = 'blib';
+  my $blib = $self->blib;
   $self->add_to_cleanup($blib);
   File::Path::mkpath( File::Spec->catdir($blib, 'arch') );
   
@@ -902,6 +1058,12 @@ sub ACTION_build {
   $self->process_script_files;
 }
 
+sub ACTION_build {
+  my $self = shift;
+  $self->depends_on('code');
+  $self->depends_on('docs');
+}
+
 sub compile_support_files {
   my $self = shift;
   my $p = $self->{properties};
@@ -931,7 +1093,10 @@ sub process_xs_files {
   my $self = shift;
   my $files = $self->find_xs_files;
   while (my ($from, $to) = each %$files) {
-    $self->copy_if_modified( from => $from, to => $to ) unless $from eq $to;
+    unless ($from eq $to) {
+      $self->add_to_cleanup($to);
+      $self->copy_if_modified( from => $from, to => $to );
+    }
     $self->process_xs($to);
   }
 }
@@ -940,7 +1105,7 @@ sub process_pod_files {
   my $self = shift;
   my $files = $self->find_pod_files;
   while (my ($file, $dest) = each %$files) {
-    $self->copy_if_modified(from => $file, to => File::Spec->catfile('blib', $dest) );
+    $self->copy_if_modified(from => $file, to => File::Spec->catfile($self->blib, $dest) );
   }
 }
 
@@ -948,7 +1113,7 @@ sub process_pm_files {
   my $self = shift;
   my $files = $self->find_pm_files;
   while (my ($file, $dest) = each %$files) {
-    $self->copy_if_modified(from => $file, to => File::Spec->catfile('blib', $dest) );
+    $self->copy_if_modified(from => $file, to => File::Spec->catfile($self->blib, $dest) );
   }
 }
 
@@ -957,7 +1122,7 @@ sub process_script_files {
   my $files = $self->find_script_files;
   return unless keys %$files;
 
-  my $script_dir = File::Spec->catdir('blib', 'script');
+  my $script_dir = File::Spec->catdir($self->blib, 'script');
   File::Path::mkpath( $script_dir );
   
   foreach my $file (keys %$files) {
@@ -1077,19 +1242,105 @@ eval 'exec $interpreter $arg -S \$0 \${1+"\$\@"}'
 }
 
 
-sub ACTION_manifypods {
+sub ACTION_docs {
   my $self = shift;
-  warn "Sorry, the 'manifypods' action is not yet implemented.\n"; return;
+  $self->depends_on('code');
   require Pod::Man;
+  $self->manify_bin_pods() if $self->install_destination('bindoc');
+  $self->manify_lib_pods() if $self->install_destination('libdoc');
+}
+
+sub manify_bin_pods {
+  my $self    = shift;
+  my $parser  = Pod::Man->new( section => 1 ); # binary manpages go in section 1
+  my $files   = $self->_find_pods($self->{properties}{bindoc_dirs});
+  return unless keys %$files;
   
-  my $p = Pod::Man->new(section => 3);
-  my $files = $self->rscan_dir('lib', qr{\.(pm|pod)$});
-  foreach my $file (@$files) {
-    my @path = File::Spec->splitdir($file);
-    # ...
+  my $mandir = File::Spec->catdir( $self->blib, 'bindoc' );
+  File::Path::mkpath( $mandir, 0, 0777 );
+
+  foreach my $file (keys %$files) {
+    my $manpage = $self->man1page_name( $file ) . '.' . $self->{config}{man1ext};
+    my $outfile = File::Spec->catfile( $mandir, $manpage);
+    next if $self->up_to_date( $file, $outfile );
+    print "Manifying $file -> $outfile\n";
+    $parser->parse_from_file( $file, $outfile );
+    $files->{$file} = $outfile;
   }
 }
 
+sub manify_lib_pods {
+  my $self    = shift;
+  my $parser  = Pod::Man->new( section => 3 ); # library manpages go in section 3
+  my $files   = $self->_find_pods($self->{properties}{libdoc_dirs});
+  return unless keys %$files;
+  
+  my $mandir = File::Spec->catdir( $self->blib, 'libdoc' );
+  File::Path::mkpath( $mandir, 0, 0777 );
+
+  foreach my $file (keys %$files) {
+    my $manpage = $self->man3page_name( $file ) . '.' . $self->{config}{man3ext};
+    my $outfile = File::Spec->catfile( $mandir, $manpage);
+    next if $self->up_to_date( $file, $outfile );
+    print "Manifying $file -> $outfile\n";
+    $parser->parse_from_file( $file, $outfile );
+    $files->{$file} = $outfile;
+  }
+}
+
+sub _find_pods {
+  my ($self, $dirs) = @_;
+  my %files;
+  foreach my $spec (@$dirs) {
+    my $dir = $self->localize_file_path($spec);
+    next unless -e $dir;
+    do { $files{$_} = $_ if $self->contains_pod( $_ ) }
+      for @{ $self->rscan_dir( $dir ) };
+  }
+  return \%files;
+}
+
+sub contains_pod {
+  my ($self, $file) = @_;
+  return '' unless -T $file;  # Only look at text files
+  
+  my $fh = IO::File->new( $file ) or die "Can't open $file: $!";
+  while (my $line = <$fh>) {
+    return 1 if $line =~ /^\=(?:head|pod|item)/;
+  }
+  
+  return '';
+}
+
+# Adapted from ExtUtils::MM_Unix
+sub man1page_name {
+  my $self = shift;
+  return File::Basename::basename( shift );
+}
+
+# Adapted from ExtUtils::MM_Unix and Pod::Man
+# Depending on M::B's dependency policy, it might make more sense to refactor
+# Pod::Man::begin_pod() to extract a name() methods, and use them...
+#    -spurkis
+sub man3page_name {
+  my $self = shift;
+  my $file = File::Spec->canonpath( shift ); # clean up file path
+  my @dirs = File::Spec->splitdir( $file );
+
+  # more clean up - assume all man3 pods are under 'blib/lib' or 'blib/arch'
+  # to avoid the complexity found in Pod::Man::begin_pod()
+  shift @dirs while ($dirs[0] =~ /^(?:blib|lib|arch)$/i);
+
+  # remove known exts from the base name
+  $dirs[-1] =~ s/\.p(?:od|m|l)\z//i;
+
+  return join( $self->manpage_separator, @dirs );
+}
+
+sub manpage_separator {
+  return '::';
+}
+
 # For systems that don't have 'diff' executable, should use Algorithm::Diff
 sub ACTION_diff {
   my $self = shift;
@@ -1099,7 +1350,7 @@ sub ACTION_diff {
   my @flags = @{$self->{args}{ARGV}};
   @flags = $self->split_like_shell($self->{args}{flags} || '') unless @flags;
   
-  my $installmap = $self->install_map('blib');
+  my $installmap = $self->install_map;
   delete $installmap->{read};
 
   my $text_suffix = qr{\.(pm|pod)$};
@@ -1135,14 +1386,14 @@ sub ACTION_install {
   my ($self) = @_;
   require ExtUtils::Install;
   $self->depends_on('build');
-  ExtUtils::Install::install($self->install_map('blib'), 1, 0, $self->{args}{uninst}||0);
+  ExtUtils::Install::install($self->install_map, 1, 0, $self->{args}{uninst}||0);
 }
 
 sub ACTION_fakeinstall {
   my ($self) = @_;
   require ExtUtils::Install;
   $self->depends_on('build');
-  ExtUtils::Install::install($self->install_map('blib'), 1, 1, $self->{args}{uninst}||0);
+  ExtUtils::Install::install($self->install_map, 1, 1, $self->{args}{uninst}||0);
 }
 
 sub ACTION_versioninstall {
@@ -1168,7 +1419,7 @@ sub ACTION_clean {
 sub ACTION_realclean {
   my ($self) = @_;
   $self->depends_on('clean');
-  $self->delete_filetree($self->{properties}{config_dir}, $self->{properties}{build_script});
+  $self->delete_filetree($self->config_dir, $self->build_script);
 }
 
 sub ACTION_ppd {
@@ -1210,10 +1461,11 @@ sub _sign_dir {
   # the right directory after a signature failure.  Would be nice if
   # Module::Signature took a directory argument.
   
+  my $start_dir = $self->cwd;
   chdir $dir or die "Can't chdir() to $dir: $!";
   eval {Module::Signature::sign()};
   my @err = $@ ? ($@) : ();
-  chdir $self->base_dir or push @err, "Can't chdir() back to " . $self->base_dir . ": $!";
+  chdir $start_dir or push @err, "Can't chdir() back to $start_dir: $!";
   die join "\n", @err if @err;
 }
 
@@ -1257,7 +1509,7 @@ sub ACTION_distdir {
   my $dist_dir = $self->dist_dir;
   $self->delete_filetree($dist_dir);
   $self->add_to_cleanup($dist_dir);
-  ExtUtils::Manifest::manicopy($dist_files, $dist_dir, 'best');
+  ExtUtils::Manifest::manicopy($dist_files, $dist_dir, 'cp');
   warn "*** Did you forget to add $self->{metafile} to the MANIFEST?\n" unless exists $dist_files->{$self->{metafile}};
   
   $self->_sign_dir($dist_dir) if $self->{properties}{sign};
@@ -1268,14 +1520,16 @@ sub ACTION_disttest {
 
   $self->depends_on('distdir');
 
+  my $start_dir = $self->cwd;
   my $dist_dir = $self->dist_dir;
   chdir $dist_dir or die "Cannot chdir to $dist_dir: $!";
   # XXX could be different names for scripts
-  # XXX doesn't propagate @INC
+  
+  local $ENV{'PERL5LIB'} = join $self->{config}{path_sep}, @INC;
   $self->run_perl_script('Build.PL') or die "Error executing 'Build.PL' in dist directory: $!";
   $self->run_perl_script('Build') or die "Error executing 'Build' in dist directory: $!";
   $self->run_perl_script('Build', [], ['test']) or die "Error executing 'Build test' in dist directory";
-  chdir $self->base_dir;
+  chdir $start_dir;
 }
 
 sub ACTION_manifest {
@@ -1298,7 +1552,7 @@ sub script_files {
   }
   return $self->{properties}{script_files};
 }
-*scripts = \&script_files;
+BEGIN { *scripts = \&script_files; }
 
 sub valid_licenses {
   return { map {$_, 1} qw(perl gpl artistic lgpl bsd open_source unrestricted restrictive unknown) };
@@ -1343,6 +1597,9 @@ sub ACTION_distmeta {
     };
 
   $node->{generated_by} = "Module::Build version " . Module::Build->VERSION;
+  
+  # If we're in the distdir, the metafile may exist and be non-writable.
+  $self->delete_filetree($self->{metafile});
 
   # YAML API changed after version 0.30
   my $yaml_sub = $YAML::VERSION le '0.30' ? \&YAML::StoreFile : \&YAML::DumpFile;
@@ -1385,10 +1642,11 @@ sub _packages_inside {
   # XXX this SUCKS SUCKS SUCKS!  Damn you perl!
   my ($self, $file) = @_;
   my $fh = IO::File->new($file) or die "Can't read $file: $!";
-  my @packages;
-  while (defined(my $line = <$fh>)) {
-    push @packages, $1 if $line =~ /^[\s\{;]*package\s+([\w:]+)/;
-  }
+  
+  my (@packages, $p);
+  push @packages, $p while (undef, $p) = 
+    $self->_next_code_line($fh, qr/^[\s\{;]*package\s+([\w:]+)/);
+  
   return @packages;
 }
 
@@ -1434,6 +1692,7 @@ sub install_types {
 
 sub install_map {
   my ($self, $blib) = @_;
+  $blib ||= $self->blib;
 
   my %map;
   foreach my $type ($self->install_types) {
@@ -1461,7 +1720,7 @@ sub install_map {
 sub depends_on {
   my $self = shift;
   foreach my $action (@_) {
-    $self->dispatch($action);
+    $self->_call_action($action);
   }
 }
 
@@ -1562,8 +1821,9 @@ sub compile_xs {
     
     my $typemap =  $self->find_module_by_name('ExtUtils::typemap', \@INC);
     my $cf = $self->{config};
+    my $perl = $self->{properties}{perl};
     
-    my $command = (qq{$^X "-I$cf->{installarchlib}" "-I$cf->{installprivlib}" "$xsubpp" -noprototypes } .
+    my $command = (qq{$perl "-I$cf->{installarchlib}" "-I$cf->{installprivlib}" "$xsubpp" -noprototypes } .
 		   qq{-typemap "$typemap" "$file"});
     
     print $command;
@@ -1634,7 +1894,7 @@ sub process_xs {
   my $archdir;
   {
     my @dirs = File::Spec->splitdir($file_base);
-    $archdir = File::Spec->catdir('blib','arch','auto', @dirs[1..$#dirs]);
+    $archdir = File::Spec->catdir($self->blib,'arch','auto', @dirs[1..$#dirs]);
   }
   
   # .xs -> .bs
